#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: prereceivecli.py
#
# Copyright 2019 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for prereceivecli.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import argparse
import json
import logging
import logging.config
import os
from dataclasses import dataclass

import boto3
import botocore

from .configuration import HASHES_SCHEMA
from .lib import (get_project,
                  get_table_for_project_group,
                  send_slack_message,
                  HashChecker,
                  SecurityEntry,
                  parse_hook_input)


@dataclass
class AwsCredentials:
    """Stores AWS Credentials."""

    access_key_id: str
    secret_access_key: str
    session_token: str


__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''26-02-2019'''
__copyright__ = '''Copyright 2019, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos", "Alberto Rodriguez Garcia"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''prereceivecli'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())

LOGGERS_TO_DISABLE = ['botocore.credentials']


def get_arguments():
    """Gets us the cli arguments.

    Returns the args as parsed from the argsparser.
    """
    # https://docs.python.org/3/library/argparse.html
    parser = argparse.ArgumentParser(description=('A cli that implements a git server side pre-receive hook that gets '
                                                  'driven from dynamodb and reports to slack offending pushes.'))
    excluding_key_arn = parser.add_mutually_exclusive_group(required=True)
    excluding_secret_token = parser.add_mutually_exclusive_group()
    excluding_key_token = parser.add_mutually_exclusive_group()
    excluding_secret_arn = parser.add_mutually_exclusive_group()
    key = excluding_key_arn.add_argument('--key', '-k',
                                         dest='access_key',
                                         action='store',
                                         help='The aws access key',
                                         type=str,
                                         required=False)
    arn = excluding_key_arn.add_argument('--arn', '-arn',
                                         dest='aws_role_arn',
                                         action='store',
                                         help='The aws role arn, defaults to environment variable',
                                         type=str,
                                         required=False)
    token = excluding_secret_token.add_argument('--token', '-token',
                                                dest='aws_web_identity_token_file',
                                                action='store',
                                                help='The aws web identity token file, defaults to environment variable',
                                                type=argparse.FileType('r'),
                                                required=False)
    secret = excluding_secret_token.add_argument('--secret', '-s',
                                                 dest='secret_key',
                                                 action='store',
                                                 help='The aws secret key',
                                                 type=str,
                                                 required=False)
    # See: https://bugs.python.org/issue10984#msg219660
    excluding_key_token._group_actions.extend([key, token])  # pylint: disable=protected-access
    excluding_secret_arn._group_actions.extend([arn, secret])  # pylint: disable=protected-access
    parser.add_argument('--log-config',
                        '-l',
                        action='store',
                        dest='logger_config',
                        help='The location of the logging config json file',
                        default='')
    parser.add_argument('--log-level',
                        '-L',
                        help='Provide the log level. Defaults to info.',
                        dest='log_level',
                        action='store',
                        default='info',
                        choices=['debug',
                                 'info',
                                 'warning',
                                 'error',
                                 'critical'])
    parser.add_argument('--slack-web-hook', '-w',
                        dest='web_hook',
                        action='store',
                        help='The slack web_hook to post messages to',
                        type=str,
                        required=True)
    parser.add_argument('--region', '-r',
                        dest='region',
                        action='store',
                        help='The aws region to use',
                        type=str,
                        required=True)
    feature_parser = parser.add_mutually_exclusive_group(required=False)
    feature_parser.add_argument('--aggressive-check', '-a',
                                dest='aggressive',
                                action='store_true',
                                help=('Flag noting whether the project should be rejected if no entry in the security '
                                      'table'))
    feature_parser.add_argument('--no-aggressive-check', '-n',
                                dest='aggressive',
                                action='store_false',
                                help=('Flag noting whether the project should be rejected if no entry in the security '
                                      'table'))
    parser.set_defaults(aggressive=False)
    args = parser.parse_args()
    if all([args.access_key, args.secret_key is None]):
        parser.error("--key requires --secret.")
    elif all([args.aws_role_arn, args.aws_web_identity_token_file is None]):
        parser.error("--arn requires --token.")
    return args


def setup_logging(level, config_file=None):
    """Sets up the logging.

    Args:
        level: The level to log for.
        config_file: The config file with the logging configuration. If provided it superseeds the level arg.

    Returns:
        args: The parsed arguments.

    """
    # This will configure the logging, if the user has set a config file.
    # If there's no config file, logging will default to stdout.
    if config_file:
        # Get the config for the logger. Of course this needs exception
        # catching in case the file is not there and everything. Proper IO
        # handling is not shown here.
        configuration = json.loads(open(config_file).read())
        # Configure the logger
        logging.config.dictConfig(configuration)
    else:
        handler = logging.StreamHandler()
        handler.setLevel(level.upper())
        formatter = logging.Formatter(('%(asctime)s - '
                                       '%(name)s - '
                                       '%(levelname)s - '
                                       '%(message)s'))
        handler.setFormatter(formatter)
        LOGGER.addHandler(handler)
        LOGGER.setLevel(level.upper())
    for logger in LOGGERS_TO_DISABLE:
        logging.getLogger(logger).disabled = True


def get_credentials(args):
    """Gets AWS credentials.

    Needs the args to either assume role or get credentials

    Credentials:
        Credentials: The AWS credentials to set for our environment

    """
    if not all([args.access_key, args.secret_key]):
        client = boto3.client('sts',
                              aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID'),
                              aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY'),
                              aws_session_token=os.environ.get('AWS_SESSION_TOKEN'))
        with open(args.aws_web_identity_token_file.name, 'r') as opened_file:
            token = opened_file.read()
        try:
            response = client.assume_role_with_web_identity(RoleArn=args.aws_role_arn,
                                                            RoleSessionName='prereceive',
                                                            WebIdentityToken=token)
        except botocore.exceptions.ClientError:
            LOGGER.exception('client')
        return AwsCredentials(response['Credentials']['AccessKeyId'],
                              response['Credentials']['SecretAccessKey'],
                              response['Credentials']['SessionToken'])
    return AwsCredentials(args.access_key,
                          args.secret_key,
                          None)


def validate_commit(project, dynamodb_table, web_hook, aggressive_checking):
    """Validates that no unauthorized change has been performed on protected files on a specified commit.

    Args:
        project (Project): An object exposing attributes of the required variables.
        dynamodb_table (Table): The dynamodb table with the entries for the projects.
        web_hook (str): The url of the slack webhook.
        aggressive_checking (bool): If set any unmatched repositories will be rejected.

    Returns:
        success (bool): True if the commit is valid False otherwise.

    """
    entries = dynamodb_table.get_item(Key={'slug': project.slug}).get('Item', {})
    if aggressive_checking and not entries:
        message = f'No dynamodb entries found for project "{project.slug}", project rejected due to aggressive checking'
        LOGGER.info(message)
        # This print is required to provide feedback to the user through the git hook via stdout
        print(message)
        return False
    if not entries:
        LOGGER.info('No dynamodb entries found for project "%s", project not secured', project.slug)
        return True
    entries = HASHES_SCHEMA.validate(entries)
    entries = [SecurityEntry(entry.get('hashes', []),
                             entry.get('name', ''),
                             entry.get('type', '')) for entry in entries.get('protected_items', [])]
    if not any([entry.hashes for entry in entries]):
        LOGGER.info('No hashes found for project "%s" for any type, project not secured', project.slug)
        return True
    errors = HashChecker().verify(project, entries)
    success = False if errors else True  # pylint: disable=simplifiable-if-expression
    for error_message in errors:
        send_slack_message(web_hook, error_message)
    return success


def main():
    """Main method.

    This method holds what you want to execute when
    the script is run on command line.

    """
    try:
        args = get_arguments()
        setup_logging(args.log_level, args.logger_config)
        # get the base, commit and ref as provided by the calling pre-receive hook
        base, commit = parse_hook_input()
        project = get_project(base, commit)
        if not project.group:
            LOGGER.info('Action seems as automatic merge for project "%s" from user "%s" and not a pre-receive hook, '
                        'letting through', project.slug, project.username)
            raise SystemExit(0)
        if not project.commit:
            LOGGER.info('It seems only tags were pushed for project "%s" by username "%s", letting through',
                        project.slug, project.username)
            raise SystemExit(0)
        os.environ['AWS_DEFAULT_REGION'] = args.region
        dynamodb_table = get_table_for_project_group(project.group, get_credentials(args))
        if not dynamodb_table:
            LOGGER.info('Project "%s" does not appear to have security settings set, letting through...', project.slug)
            raise SystemExit(0)
        success = validate_commit(project, dynamodb_table, args.web_hook, args.aggressive)
    except Exception:
        LOGGER.exception('Some unexpected error occurred letting things through!')
        raise SystemExit(0)
    raise SystemExit(0 if success else 1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File: configuration.py
#
# Copyright 2019 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for configuration.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import logging
import re
from schema import Schema

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-01-2019'''
__copyright__ = '''Copyright 2019, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''configuration'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())

# SLUG_REGEX = re.compile(r'^[0-9]{3,}-[a-z0-9-]+')
SLUG_REGEX = re.compile(r'.*')

HASHES_SCHEMA = Schema({u'protected_items': [{u'hashes': [str],
                                              u'name': str,
                                              u'type': lambda x: x in ['file', 'directory']}],
                        u'slug': SLUG_REGEX.match})

ERROR_MESSAGE = (':fire:'
                 'Hello, *{project.username}* - you are not allowed to change {entry.type} *{entry.name}*.'
                 ' As a result this server is not accepting your push.\n'
                 'Calculated hash for {entry.type} *{entry.name}* was: *{calculated_hash}*'
                 '{diff}'
                 'If you want this commit to be accepted please contact the responsible team, '
                 'or add the hash to the appropriate entry in '
                 '<https://{region}.console.aws.amazon.com/dynamodb/home?region={region}#tables:'
                 'selected={project.group}_git_hook;tab=items|DynamoDB>')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File: utils.py
#
# Copyright 2019 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for utils.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import logging
import json
import os
import pathlib
import sys
import shlex
import tempfile
import zipfile
from dataclasses import dataclass
from subprocess import Popen, PIPE, check_output, CalledProcessError

import boto3
import requests
from botocore.exceptions import NoRegionError, NoCredentialsError, ClientError
from commonutilslib import Hasher, Pushd
from prereceivecli.configuration import ERROR_MESSAGE
from prereceivecli.prereceivecliexceptions import GitExecutionPathNotFound

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-01-2019'''
__copyright__ = '''Copyright 2019, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos", "Sayantan Khanra", "Ninad Page"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''utils'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())


@dataclass
class Project:
    """Models a project exposing attributes for slug, group and git_path."""

    slug: str
    group: str
    git_path: str
    git_command: str
    username: str
    commit: str
    base: str


@dataclass
class SecurityEntry:
    """Models a security entry exposing attributes for slug, type and git_path."""

    hashes: list
    name: str
    type: str


def parse_hook_input():
    """Parses the git hook input disregarding the tags reference."""
    hook_input = sys.stdin.read()
    base = None
    commit = None
    try:
        for line in hook_input.splitlines():
            temp_base, temp_commit, refs = line.strip().split()
            if refs.strip().startswith('refs/tags/'):
                LOGGER.info('Disregarding tag push with base: %s and commit: %s for tag %s',
                            temp_base, temp_commit, refs)
                continue
            base = temp_base
            commit = temp_commit
    except (ValueError, TypeError, AttributeError):
        LOGGER.error('Received invalid input from hook, bailing... Value provided was : %s', hook_input)
        raise SystemExit(0)
    return base, commit


def execute_command_with_returned_output(command):
    """Execute the command with returned output."""
    stdout = ''
    stderr = ''
    command = shlex.split(command)
    try:
        LOGGER.debug('Executing command %s', command)
        command_execution = check_output(command)
        stdout = command_execution.decode('utf-8')
    except CalledProcessError as command_execution:
        stderr = command_execution.stderr.decode('utf-8')
    success = bool(command_execution)
    return success, stdout.strip(), stderr.strip()


def send_slack_message(webhook, message):
    """Send a message to a webhook in slack.

    Args:
        webhook (str): The webhook to submit the message to
        message (str): The message to submit to slack

    Returns:
        (bool): True on success False otherwise

    """
    slack_data = {'text': f'{message}'}
    success = True
    if not webhook:
        LOGGER.warning('No web slack web hook provided, not reporting to slack')
        return success
    try:
        response = requests.post(webhook,
                                 data=json.dumps(slack_data),
                                 headers={'Content-Type': 'application/json'})
        if not response.ok:
            LOGGER.error('Request to slack returned an error %s, the response is:\n%s',
                         response.status_code, response.text)
    except Exception:  # pylint: disable=broad-except
        success = False
        LOGGER.exception('Problem encountered sending slack message.')
    return success


def get_project(base, commit):
    """Constructs a project object from a gitlab project path.

    Returns:
        (project): An object exposing the required attributes of the environment and the project

    """
    username = os.environ.get('GL_USERNAME')
    git_execution_path = os.environ.get('GIT_EXEC_PATH')
    if git_execution_path:
        git_command = f'{git_execution_path}/git'
    else:
        success, stdout, _ = execute_command_with_returned_output('git --exec-path')
        git_command = f'{stdout}/git' if success else None
    if not git_command:
        raise GitExecutionPathNotFound()
    git_path = os.environ.get('PWD')
    components = os.environ.get('GL_PROJECT_PATH').split('/')
    project_slug = components[-1]
    project_group = '/'.join(components[:-1])
    return Project(project_slug, project_group, git_path, git_command, username, commit, base)


def get_table_for_project_group(project_group, credentials):
    """Retrieves a dynamodb table following a specific naming convention.

    Args:
        project_group (str): The type of the project to look up the table for.
            Convention states that the table should be named {type, eg:infrastructure}_git_hook.
        credentials (AwsCredentials): An object holding the credentials passed from the authentication process.

    Returns:
        (dynamodb Table): if found else None

    """
    invalid_settings = 'Invalid aws credentials settings. Please set region and credentials properly'
    try:
        dynamodb = boto3.resource('dynamodb',
                                  aws_access_key_id=credentials.access_key_id,
                                  aws_secret_access_key=credentials.secret_access_key,
                                  aws_session_token=credentials.session_token)
    except NoRegionError:
        LOGGER.exception('')
        raise ValueError(invalid_settings)
    project_group = project_group.replace('/', '_').replace('-', '_')
    table = dynamodb.Table(f'{project_group}_git_hook')
    try:
        table.item_count
    except ClientError:
        LOGGER.debug('No table found for project group %s', project_group)
        return None
    except NoCredentialsError:
        LOGGER.error(invalid_settings)
        raise ValueError(invalid_settings)
    return table


class HashChecker:  # pylint: disable=too-few-public-methods
    """Implements a git rebuilding context manager for a pre-receive hook."""

    def __init__(self):
        self._logger = logging.getLogger(f'{LOGGER_BASENAME}.{self.__class__.__name__}')

    def _execute_command(self, command):
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        out, err = process.communicate()
        if process.returncode:
            self._logger.error('Error executing command "%s"\n\tstderr: "%s"\n\tstdout: "%s"',
                               command, err, out)
        return True if not process.returncode else False  # pylint: disable=simplifiable-if-expression

    def _export_revision(self, project, archive_path, extract_path):
        errors = []
        self._logger.info('Creating an archive from "%s" @rev "%s" to "%s"',
                          project.git_path, project.commit, archive_path)
        if not self._execute_command([project.git_command,
                                      'archive',
                                      '--format=zip', '-o', f'{archive_path}', '-0',
                                      f'{project.commit}']):
            error_message = f'Failed creating an archive from {project.commit}'
            self._logger.error(error_message)
            errors.append(error_message)
            return errors
        self._logger.info('Extracting archive "%s" to "%s"', archive_path, extract_path)
        with zipfile.ZipFile(archive_path, 'r') as zip_ref:
            zip_ref.extractall(extract_path)
        return errors

    def _validate_hashes(self, project, entries, extract_path):
        errors = []
        hasher = Hasher()
        self._logger.info('Changing directory to "%s"', extract_path)
        with Pushd(extract_path):
            for entry in entries:
                if not entry.hashes:
                    self._logger.info('"%s": No hashes are set, skipping check for %s %s',
                                      project.slug, entry.type, entry.name)
                    continue
                path = str(pathlib.Path(extract_path, entry.name).resolve())
                self._logger.info('"%s": Calculating hash for "%s" "%s" in path: "%s"',
                                  project.slug, entry.type, entry.name, path)
                calculated_hash = getattr(hasher, f'hash_{entry.type}')(path)
                if calculated_hash not in entry.hashes:
                    process = Popen([project.git_command, 'diff', project.base,
                                     project.commit], stdout=PIPE, stderr=PIPE)
                    out, _ = process.communicate()
                    error_message = (ERROR_MESSAGE.format(entry=entry,
                                                          project=project,
                                                          calculated_hash=calculated_hash,
                                                          diff=self._render_diff(
                                                              out.decode("utf-8"),
                                                              entry.name),
                                                          region=os.environ[
                                                              'AWS_DEFAULT_REGION']))
                    errors.append(error_message)
                    self._logger.error(error_message)
        return errors

    def _render_diff(self, text, file_name):
        text = self._filter_diff_entry(text, file_name)
        if text:
            output = (f'\nOffending diff is :\n'
                      f'```{text}```\n')
        else:
            output = '\nNo diff available.\n'
        return output

    def _filter_diff_entry(self, diff, file_name):
        diff = diff.replace('\ndiff', '\n|||diff')
        try:
            text = r''.join([line for line in diff.split('|||')
                             if file_name in line.splitlines()[0]])
        except IndexError:
            self._logger.warning('No diff found')
            text = ''
        return text

    def verify(self, project, entries):
        """Verifies the protected files or directories specified in the entries.

        Args:
            project (Project): The project object to verify.
            entries (dict): The entries of protected files or directories.

        Returns:
            errors (list): A list of errors of the verification failures if any.

        """
        # `GIT_DIR` is unchanged, so we are operating on the main (server's) copy of the repo
        # DO NOT DO ANYTHING THAT WILL UPDATE THE REFS (such as a git checkout) as that might
        # corrupt the repo
        with tempfile.TemporaryDirectory() as temporary_directory:
            self._logger.info('Created a temporary directory "%s"', temporary_directory)
            # `TemporaryDirectory()` does not change cwd, so we are still in the project's GIT_DIR
            archive_path = str(pathlib.Path(temporary_directory, 'archive.zip').resolve())
            extract_path = str(pathlib.Path(temporary_directory, 'archive').resolve())
            errors = self._export_revision(project, archive_path, extract_path)
            if errors:
                return errors
            errors = self._validate_hashes(project, entries, extract_path)
        self._logger.info('Cleaning up temporary directory "%s"', temporary_directory)
        return errors
